!*****************************************************************
! Module for solving scalar relativistic radial equations
!   Uses formalism of Shadwick, Talman, and Norman, Comp. Phys. Comm.
!      54, 95-102 (1989)
!      Use Numerov algorithm to integrate second order equation for upper
!       radial component;   averaged over kappa for given l
!   06-16-06  NAWH
!*****************************************************************    
MODULE radialsr
  USE GlobalMath
  USE gridmod
  USE atomdata
  USE calcpotential

  IMPLICIT NONE

  Real(8), private, parameter :: starttol=1.d-4
  Real(8), private :: gamma,c1,c2
  Real(8), private, allocatable :: ww(:),jj(:)
     ! for y''=Ay; ww stores A(r)
     ! jj stores (r+(alpha/2)**2*(E*r-rv)

CONTAINS

!******************************************************************
! Allocate_scalar_relativistic
!******************************************************************
 Subroutine Allocate_scalar_relativistic(Grid)
   Type(GridInfo), INTENT(IN) :: Grid

   INTEGER :: n,i

   n=Grid%n
   allocate(ww(n),jj(n), stat=i)
        if (i/=0) then
           write(6,*)  'Allocate_scalar_relativistic: error in allocation ',&
                       i,n
           stop
        endif

 End subroutine Allocate_scalar_relativistic

!******************************************************************
! Deallocate_scalar_relativistic
!******************************************************************
 Subroutine deallocate_scalar_relativistic

    deallocate(ww,jj)

 end subroutine deallocate_scalar_relativistic

!*******************************************************************
!  Subroutine Azeroexpand(Grid,Pot,l,energy)
!*******************************************************************
 Subroutine Azeroexpand(Grid,Pot,l,energy,nr)
   Type(GridInfo), INTENT(IN) :: Grid
   Type(PotentialInfo), INTENT(IN) :: Pot
   Integer, INTENT(IN) :: l
   Real(8), INTENT(IN) :: energy
   Integer, optional, INTENT(IN) :: nr

   Integer :: i,j,k,n,nz
   Real(8) :: xx,yy,angm,alpha2,balpha2
   Real(8), allocatable :: drv(:),ddrv(:)

   n=Grid%n
   if (present(nr)) n=min(n,nr)

   nz=Pot%nz
   allocate(drv(n),ddrv(n), stat=i)
      if (i/=0) then
         write(6,*) 'Azeroexpand:   error in allocation ', n,i
         stop
      endif
    ww=0; jj=0;  drv=0;   ddrv=0
   balpha2=inverse_fine_structure**2
   alpha2=1.d0/balpha2
   !write(6,*) 'in Azeroexpand', alpha2,nz,n
   jj(1:n)=(Grid%r(1:n) + &
        0.25d0*alpha2*(energy*Grid%r(1:n)-Pot%rv(1:n)))
   call derivative(Grid,Pot%rv(1:n),drv(1:n),1,n)
   call derivative(Grid,drv(1:n),ddrv(1:n),1,n)
   ddrv(2:n)=0.25d0*alpha2*ddrv(2:n)/jj(2:n)
   drv(2:n)=0.25d0*alpha2*(drv(2:n)-Pot%rv/Grid%r(2:n))/jj(2:n)
   angm=l*(l+1)
   ww(2:n)=(Pot%rv(2:n)/Grid%r(2:n)-energy)*(jj(2:n)/Grid%r(2:n)) &
     +0.75d0*(drv(2:n)**2)+0.5d0*ddrv(2:n) + angm/(Grid%r(2:n)**2)
   ww(1)=0
   gamma=sqrt(angm+1.d0-alpha2*nz**2)
   c1=-((2.d0*balpha2)/(nz*(2*gamma+1)))*((1-gamma)*&
           (1+0.25d0*alpha2*(energy-Pot%v0))+alpha2*nz**2+&
           0.5d0*((nz*alpha2)**2)*(energy-Pot%v0))
   xx=2*nz*(1+alpha2*(energy-Pot%v0)+3*(0.25*alpha2*(energy-Pot%v0))**2)&
       +0.25d0*alpha2*Pot%v0p*(alpha2*nz*nz+2*(gamma-1))
   yy=(1+gamma)*(1+0.25d0*alpha2*(energy-Pot%v0))+alpha2*nz*nz+&
           0.5d0*((alpha2*nz)**2)*(energy-Pot%v0)
   c2=-(xx+yy*c1)/(2*(gamma+1)*alpha2*nz)  
   write(6,*) 'Azeroexpand: ', gamma,c1,c2
   deallocate(drv,ddrv)

    ! do i=2,n
    !   write(97,'(1p5e15.7)') Grid%r(i),Pot%rv(i)/Grid%r(i),ww(i),jj(i)/Grid%r(i)
    ! enddo

  end subroutine Azeroexpand
   
!*******************************************************************
! SUBROUTINE wfnsrinit(Grid,wfn)
!*******************************************************************
  SUBROUTINE wfnsrinit(Grid,wfn,istart,nr)
   ! returns the solution of the Schroedinger equation near r=0
   !  using power series expansion
   Type(GridInfo), INTENT(IN) :: Grid
    REAL(8),INTENT(INOUT) :: wfn(:)
    INTEGER, INTENT(OUT) :: istart
    INTEGER, OPTIONAL, INTENT(IN) :: nr

    REAL(8) :: rr
    INTEGER :: i,j,n

    wfn=0
    
    n=Grid%n
    if (present(nr)) n=MIN0(n,nr)
    write(6,*) 'in wfnsrinit', n
    istart=4
    do i=1,n-1
       rr=Grid%r(i+1)
       wfn(i+1)=1+rr*(c1+rr*c2)
       wfn(i+1)=wfn(i+1)*(rr**gamma)
       wfn(i+1)=wfn(i+1)*sqrt(rr/jj(i+1))
       if  (ABS(rr*c2/c1)>starttol.and.i>=istart) exit
    enddo

    if (i<n.and.i>istart) istart=i
    !write(6,*) 'istart = ', istart
       
  End SUBROUTINE wfnsrinit
    
  
  !**********************************************************************
  !      subroutine unboundsr(Grid,Pot,nr,l,energy,wfn,nodes)
  !  pgm to solve radial scalar relativistic equation for unbound states
  !    at energy 'energy' and at angular momentum l
  !
  !    with potential rv/r, given in uniform linear or log mesh of n points
  !   assuming p(r)=C*r**(l+1)*polynomial(r) for r==0;
  !   
  !  nz=nuclear charge
  !
  !  uses Noumerov algorithm
  !
  ! also returns node == number of nodes for calculated state
  !************************************************************************
  SUBROUTINE unboundsr(Grid,Pot,nr,l,energy,wfn,nodes)
    TYPE(GridInfo), INTENT(IN)  :: Grid
    TYPE(PotentialInfo), INTENT(IN)  :: Pot
    INTEGER, INTENT(IN) :: nr,l
    REAL(8), INTENT(IN) :: energy
    REAL(8), INTENT(INOUT) :: wfn(:)
    INTEGER, INTENT(INOUT) :: nodes

    INTEGER :: n,nz,i,j,k,ierr,istart
    REAL(8) :: scale

    n=Grid%n
    IF (nr > n) THEN
       WRITE(6,*) 'Error in unboundsr -- nr > n', nr,n
       STOP
    ENDIF

    call Azeroexpand(Grid,Pot,l,energy,nr)
    call wfnsrinit(Grid,wfn,istart,nr)

    call shifted_forward_numerov(Grid,nr,istart,ww,wfn,nodes)
    !
    ! normalize to unity within integration range
    !
    wfn(2:nr)=wfn(2:nr)*sqrt(jj(2:nr)/Grid%r(2:nr))
    wfn(1)=0
    call filter(nr,wfn,machine_zero)
    scale=1.d0/overlap(Grid,wfn(1:nr),wfn(1:nr),1,nr)
    scale=SIGN(SQRT(scale),wfn(nr-2))
    wfn(1:nr)=wfn(1:nr)*scale  


  END SUBROUTINE unboundsr

!******************************************************************
!  SUBROUTINE boundsr(Grid,Pot,Orbit,l,start,nroot,emin,ierr)
!******************************************************************
  SUBROUTINE boundsr(Grid,Pot,Orbit,l,start,nroot,emin,ierr)
    !  pgm to solve radial scalar relativistic equation for nroot bound state
    !    energies and wavefunctions for angular momentum l
    !    with potential rv/r, given in uniform linear or log mesh of n points
    !  nz=nuclear charge
    !  emin=is estimate of lowest eigenvalue; used if nz=0
    !     otherwise, set to the value of -(nz/(l+1))**2
    !
    !  It is assumed that the wavefunction has np-l-1 nodes, where
    !    np is the principle quantum number-- np=1,2,..nroot
    !
    !  uses Noumerov algorithm
    !
    !  Corrections are also needed for r>n*h, depending on:
    !         e0 (current guess of energy eigenvalue
    !         the extrapolated value of rv == r * v
    !
    ! ierr=an nroot digit number indicating status of each root
    !   a digit of 1 indicates success in converging root
    !              2 indicates near success in converging root
    !              9 indicates that root not found
    !
    ! first check how many roots expected =  ntroot (returned as argument)
    !
    TYPE(GridInfo), INTENT(IN) :: Grid
    TYPE(PotentialInfo), INTENT(IN) :: Pot
    TYPE(OrbitInfo), INTENT(INOUT) :: Orbit
    INTEGER, INTENT(IN) :: l,start,nroot
    INTEGER, INTENT(INOUT) :: ierr
    REAL(8), INTENT(INOUT) :: emin

    REAL(8), PARAMETER :: convre=1.d-10,vlrg=1.d30
    INTEGER, PARAMETER :: niter=1000

    REAL(8), POINTER :: rv(:),eig(:),wfn(:,:)
    REAL(8), ALLOCATABLE :: p1(:),p2(:),dd(:)
    INTEGER :: nz,n
    REAL(8) :: h,v0,v0p
    REAL(8) :: err,convrez,energy,zeroval
    REAL(8) :: scale,emax,best,rout,ppp
    REAL(8) :: arg,r,r2,veff,pppp1,rin,dele,x,rvp1,pnp1,bnp1
    INTEGER :: iter,i,j,k,node,match,mxroot,ntroot,ir,iroot
    INTEGER :: least,many,ifac,istart
    LOGICAL :: ok
    ! integer :: icount=0

    n=Grid%n
    h=Grid%h
    wfn=>Orbit%wfn
    eig=>Orbit%eig
    ALLOCATE(p1(n),p2(n),dd(n),stat=i)
    IF (i/=0) THEN
       WRITE(6,*) ' Error in boundsr allocation ',i,n
       STOP
    ENDIF

    nz=Pot%nz
    v0=Pot%v0
    v0p=Pot%v0p
    rv=>Pot%rv
    err=n*nz*(h**4)
    convrez=convre
    IF (nz.GT.0) convrez=convre*nz
    !     write(6,*) 'expected error = ',err
    ierr=0

    WRITE(6,*) 'z , l = ',nz,l
    ! check how many roots expected by integration outward at
    !   energy = 0
    energy = 0
    call Azeroexpand(Grid,Pot,l,energy)
    call wfnsrinit(Grid,p1,istart)
       !do i=1,istart
       !   write(6,'(1p2e15.7)') Grid%r(i),p1(i)
       !enddo


    !
    !  start outward integration

    call shifted_forward_numerov(Grid,n,istart,ww,p1,node)

    WRITE(6,*) ' nodes at e=0  ', node
       !do i=1,n
       !   write(6,'(1p4e15.7)') Grid%r(i),p1(i),jj(i),ww(i)
       !enddo
    
    mxroot=node+1
    ntroot=node
    IF (mxroot.LT.nroot) THEN
       WRITE(6,*)'error in boundsr - for l = ',l
       WRITE(6,*) nroot,' states requested but only',mxroot,' possible'
       DO ir=mxroot+1,nroot
          ierr=ierr+9*(10**(ir-1))
       ENDDO
    ENDIF
    mxroot=min0(mxroot,nroot)
    !
    IF (nz.EQ.0) energy=-ABS(emin)
    IF (nz.NE.0) energy=-1.1d0*(nz/(l+1.d0))**2 
    emin=energy-err
    emax=0.d0

    DO iroot=1,mxroot
       best=1.d10; dele=1.d10
       energy=emin+err
       IF (energy.LT.emin) energy=emin
       IF (energy.GT.emax) energy=emax
       ok=.FALSE.
       !write(6,*) 'iter,iroot,energy',iter,iroot,energy
       !write(6,*) 'emin,max',emin,emax
       BigIter: DO iter=1,niter
          !write(6,*) 'In iter with energy', iter,energy,niter,l,iroot
          !  start inward integration
          !  start integration at n
          call Azeroexpand(Grid,Pot,l,energy)
          ! find classical turning point
          least=min0(10,n/4)
          call wfnsrinit(Grid,p1,istart)
          least=max0(istart,least)       ! find istart, initial p1
          call scalarrelativisticturningpt(Grid,least,match)
          x=0.5d0*(ww(n)+ww(n-1))
          ppp=SQRT(ABS(x-energy))
          p2=0
          p2(n)=1
          p2(n-1)=exp(-ppp*(Grid%r(n-1)-Grid%r(n)))
          call mod_backward_numerov(Grid,match,ww,p2)
          p2(match:n)=p2(match:n)*sqrt(jj(match:n)/Grid%r(match:n))
          match=match+6
          rin=Gfirstderiv(Grid,match,p2)/p2(match)

          call shifted_forward_numerov(Grid,match+6,istart,ww,p1,node)
          p1(2:match+6)=p1(2:match+6)*sqrt(jj(2:match+6)/Grid%r(2:match+6))
          p1(1)=0
          !icount=icount+1
          !  do i=1,match+6
          !    write(100+icount,'(1P2e15.7)') Grid%r(i),p1(i)
          !  enddo
          rout=Gfirstderiv(Grid,match,p1)/p1(match)
             !write(6,*) 'node,match,rin,rout',node,(iroot-1),match,rin,rout
          ! check whether node = (iroot-1)
          !   not enough nodes -- raise energy
          IF (node.LT.iroot-1) THEN
             emin=MAX(emin,energy)-err
             energy=emax-(emax-energy)*ranx()
             ifac=9
             !   too many nodes -- lower energy
          ELSEIF (node.GT.iroot-1) THEN
             IF (energy.LE.emin) THEN
                ierr=ierr+9*(10**(iroot-1))
                WRITE(6,*) 'boundsr error -- emin too high',l,nz,emin,energy
                do i=2,n
                   write(999,'(1p4e15.7)') Grid%r(i),jj(i)/Grid%r(i),ww(i),Pot%rv(i)
                enddo
                STOP
             ENDIF
             emax=MIN(emax,energy+err)
             energy=emin+(energy-emin)*ranx()
             !   correct number of nodes -- estimate correction
          ELSEIF (node.EQ.iroot-1) THEN
             DO j=1,match
                p1(j)=p1(j)/p1(match)
                       !  write(6,*) 'j,p1',j,p1(j)
             ENDDO
             DO j=match,n
                p1(j)=p2(j)/p2(match)
                        ! write(6,*) 'j,p2',j,p1(j)
             ENDDO
             scale=1.d0/overlap(Grid,p1,p1)
             dele=(rout-rin)*scale
                  !write(6,*) 'energy,dele,scale',energy,dele,scale
             x=ABS(dele)
             IF (x.LT.best) THEN
                scale=SQRT(scale)
                p1(1:n)=p1(1:n)*scale 
                k=start+iroot-1
                call filter(n,p1,machine_zero)
                wfn(1:n,k)=p1(1:n) 
                eig(k)=energy
                !write(6,*) 'root',l,iroot,eig(k),emin,emax
                best=x
             ENDIF
             IF (ABS(dele).LE.convrez) THEN
                !write(6,*) 'iter with dele' , iter,dele
                ok=.TRUE.
                !  eigenvalue found
                ierr=ierr+10**(iroot-1)
                IF (iroot+1.LE.mxroot) THEN
                   emin=energy+err
                   emax=0
                   energy=(emin+emax)/2
                   IF (energy.LT.emin) energy=emin
                   IF (energy.GT.emax) energy=emax
                   best=1.d10
                ENDIF
                EXIT BigIter
             ENDIF
             IF (ABS(dele).GT.convrez) THEN
                !write(6,*) 'iter with dele' , iter,dele
                energy=energy+dele
                ! if energy is out of range, pick random energy in correct range
                IF (emin-energy.GT.convrez.OR.energy-emax.GT.convrez)         &
                     energy=emin+(emax-emin)*ranx()
                ifac=2
                !write(6,*) 'continuing with iter dele', iter,dele
             ENDIF
          ENDIF
       ENDDO BigIter !iter
       IF (.NOT.ok) THEN
          ierr=ierr+ifac*(10**(iroot-1))
          WRITE(6,*) 'no convergence in boundsr',iroot,l,dele,energy
          WRITE(6,*) ' best guess of eig, dele = ',eig(iroot),best
          IF (iroot.LT.mxroot) THEN
             DO ir=iroot+1,mxroot
                ierr=ierr+9*(10**(ir-1))
             ENDDO
          ENDIF
       ENDIF
    ENDDO !iroot

    ! icount=icount+1
    ! do i=1,n
    !   write(100+icount,'(1p25e15.6)') Grid%r(i),(wfn(i,j),j=start,start+nroot)
    ! enddo
    DEALLOCATE(p1,p2,dd)
             write(6,*) 'returning from boundsr -- ierr=',ierr
  END SUBROUTINE Boundsr

  subroutine scalarrelativisticturningpt(Grid,least,turningpoint)
     Type(GridInfo), INTENT(IN) :: Grid
     Integer, INTENT(IN) :: least
     Integer, INTENT(OUT) :: turningpoint

    integer :: i,n

    n=Grid%n

    turningpoint=n
    do i=n,least,-1
      if (ww(i)<0.d0) exit
    enddo
    turningpoint=i

    !write(6,*) 'Found turning point at ', turningpoint, Grid%r(turningpoint)

  End subroutine scalarrelativisticturningpt


END MODULE radialsr
